/***********************************************************/
/* Author: Alex Thomson                                    */
/* File: client.c                                          */
/* Purpose: This program acts as a RPSLS client and needs  */
/*          to be paired with the RPSLS server. The user   */
/*          inputs data and it is then transmitted to the  */
/*          server and processed to determine winners.     */
/*                                                         */
/* Networking code taken from Molay,                       */
/* "Understanding Unix/Linux Programming"                  */
/* http://wps.prenhall.com/esm_molay_UNIXProg_1/           */
/***********************************************************/

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>

/* Defines */
#define PORTNUM 25000 //The default port of the game
#define HOSTLEN  256 //The default buffer size for all unknown strings
#define TRUE  1 //If you can't figure this one out...
#define FALSE  0 //Then you don't understand this either
#define oops(msg){perror(msg);exit(1);} //Built in error function


/* Function Declarations */

int getPlayerName(char *);
void processRequest(FILE *) ;
void playARound(FILE *);
void getPlay(FILE *);
void processRound(FILE *socketFile,int win);
void processOpponent(FILE *);
int connectToServer(char *, int );
void cleanCRLF( char *);

/* Global Variable Declaration*/
char *items[] = {"","Rock","Paper","Scissors","Lizard","Spock"}; //Who array
char *actions[] = {"","smashes","cuts","covers","crushes","poisons", //What array
                    "decapitates","disproves","vaporizes","eats"};


/***********************************************************/
/* Function: main                                          */
/* Purpose: sets up the terminal and starts the game       */
/* Parameters: none (void)                                 */
/* Returns: int      exit code                             */
/***********************************************************/
int main(int argc, char **argv){
    int socketDescriptor; //the file descriptor for the socket
    FILE *socketFile; //the file of the socket
    char hostess[HOSTLEN]; //the game host
    char player[HOSTLEN]; //the player name
    int nameLen; //length of the player name

    //Pre-clean the strings
    memset(hostess,'\0',HOSTLEN);
    memset(player,'\0',HOSTLEN);

    printf("What is the RPSLS host?\n");
    fgets(hostess,HOSTLEN,stdin); //get the host name
    cleanCRLF(hostess); //strip the new line from the host name

    //Connect to the server and open the socket file or break
    socketDescriptor = connectToServer(hostess, PORTNUM);
    if (socketDescriptor == -1) oops("broke");
    socketFile = fdopen(socketDescriptor, "r+");

    //Get the player's name and it's length
    printf("What is your name?\n");
    nameLen = getPlayerName(player);

    //Write the player's name to the server
    fwrite (player , 1,  nameLen , socketFile );
    fflush(socketFile);

    // Loop forever while processing server requests
    while(TRUE){
        processRequest(socketFile);
    }

    return 0; //Just 'cuz...
}

/***********************************************************/
/* Function: processRequest                                */
/* Purpose: reads the socket file for input and directs it */
/* Parameters: FILE *socketFile - the socket file          */
/* Returns: void (none)                                    */
/***********************************************************/
void processRequest(FILE *socketFile) {
    char  inst[5]; //the instruction of what to do

    //Pre-clean the strings
    memset(inst,'\0',5);

    //Get the instruction from the server
    fread (inst,1,4,socketFile);


    if ( !strcmp(inst,"WAIT")){ //If the instruction is WAIT
        printf("Waiting for another player.\n"); //Let the player know
    }
    else if ( !strcmp(inst,"PLAY") ){//If the instruction is PLAY
        playARound(socketFile); //Play a round
    }
    else if ( !strcmp(inst,"WING") ){//If the instruction is WING
        printf("You have won the game!\n"); //the player won the game
        printf("Bye!\n");
        fclose(socketFile); //close up shop
        exit(0); //exit
    }
    else if ( !strcmp(inst,"LOSG") ){//If the instruction is LOSG
        printf("You have lost the game!\n"); //the player lost the game
        printf("Bye!\n");
        fclose(socketFile); //close up shop
        exit(0); //exit
    }
    else if ( !strcmp(inst,"WINR") ){//If the instruction is WINR
        processRound(socketFile,TRUE); //process the round as a win

    }
    else if ( !strcmp(inst,"LOSR") ){//If the instruction is LOSR
        processRound(socketFile,FALSE); //process the round as a loss

    }
    else if ( !strcmp(inst,"GETN") ){//If the instruction is GETN
        processOpponent(socketFile); //process the opponent's name
    }
}

/***********************************************************/
/* Function: processOpponent                               */
/* Purpose: sends the opponent's name to the screen        */
/* Parameters: FILE *socketFile - the socket file          */
/* Returns: void (none)                                    */
/***********************************************************/
void processOpponent(FILE *socketFile){
    char name[HOSTLEN]; //the opponent's name
    char len[4]; //the length of the name

    //Pre-clean the strings
    memset(len,'\0',4);
    memset(name,'\0',HOSTLEN);

    fread (len,1,3,socketFile);//Get the length of the name
    fread (name,1,atoi(len),socketFile); //Get the name

    printf("You are playing against %s\n",name); //Tell the player the name
}

/***********************************************************/
/* Function: processRound                                  */
/* Purpose: determines what to print for the round winner  */
/* Parameters: FILE *socketFile - the socket file          */
/*             int win - boolean for who won the game      */
/* Returns: void (none)                                    */
/***********************************************************/
void processRound(FILE *socketFile,int win){
    char winner[2],loser[2],action[2]; //the different response codes
    char winStr[HOSTLEN]; //the final win string to be printed

    //Pre-clean the strings
    memset(winner,'\0',2);memset(loser,'\0',2);memset(action,'\0',2);
    memset(winStr,'\0',HOSTLEN);

    fread (winner,1,1,socketFile); //get the winner
    fread (action,1,1,socketFile); //get the action
    fread (loser,1,1,socketFile); //get the loser

    if (!atoi(action)){ //if it was a tie, print that
        printf("It was a tie!\n");
        printf("************************************\n");
        return;
    }

    //format the action string to print
    sprintf(winStr,"%s %s %s",
            items[atoi(winner)],
            actions[atoi(action)],
            items[atoi(loser)]);

    //print about the winnings (or losings)
    (win)?(printf("You won the round!\n")):(printf("You lost the round!\n"));
    printf("%s\n",winStr);
    printf("************************************\n");
}

/***********************************************************/
/* Function: playARound                                    */
/* Purpose: play a round of RPSLS                          */
/* Parameters: FILE *socketFile - the socket file          */
/* Returns: void (none)                                    */
/***********************************************************/
void playARound(FILE *socketFile){
    char arg[4]; //the round id number

    //Pre-clean the strings
    memset(arg,'\0',4);

    fread (arg,1,3,socketFile); //get the round number
    if (atoi(arg) == 1){ // if it is the first round, print the instructions
        printf("Now we can play!\n");
        printf("To exit the game, type Quit when prompted for input.\n");
        printf("Otherwise, when you are prompted, enter one of the following:\n");
        printf("Rock, Paper, Scissors, Lizard, Spock (case sensitive)\n");
        printf("************************************\n");
        printf("Round 1\n");
    }
    else{ //otherwise print the round info
        printf("Round %d\n",atoi(arg));
    }

    getPlay(socketFile); //play the round

}

/***********************************************************/
/* Function: getPlayerName                                 */
/* Purpose: get the player name from input                 */
/* Parameters: char *playerName - the player's name        */
/* Returns: void (none)                                    */
/***********************************************************/
int getPlayerName(char *playerName){
    char player[HOSTLEN-4]; //the player name
    int argLength = 0; //the length of the player name

    //Pre-clean the strings
    //memset(player,'\0',HOSTLEN-4);

    fgets(player,HOSTLEN-4,stdin); //get the player name
    cleanCRLF(player); // clean out the new line from the name

    argLength=strlen(player); //get the length of the name

    sprintf(playerName,"%03d%s",argLength,player); //build the send string

    return (argLength + 3); //return the length of the name
}

/***********************************************************/
/* Function: getPlay                                       */
/* Purpose: get the played value from the terminal         */
/* Parameters: FILE *socketFile - the socket file          */
/* Returns: void (none)                                    */
/***********************************************************/
void getPlay(FILE *socketFile){
    char playString[8]; //the formated play string
    char inst[HOSTLEN]; //the instruction received from the terminal
    int tryAgain = TRUE; //boolean for incorrect input
    int playValue; //the numerical representation of the play

    //Pre-clean the strings
    memset(playString,'\0',8);

    while(tryAgain){ //so long as there is an error

        tryAgain = FALSE; //clear the error

        fgets(inst,HOSTLEN,stdin); //get the instruction
        cleanCRLF(inst); //clean the instruction

        //determine the necessary play value
        if(!strcmp(inst,"Rock"))         {playValue = 1;}
        else if(!strcmp(inst,"Paper"))   {playValue = 2;}
        else if(!strcmp(inst,"Scissors")){playValue = 3;}
        else if(!strcmp(inst,"Lizard"))  {playValue = 4;}
        else if(!strcmp(inst,"Spock"))   {playValue = 5;}

        else if(!strcmp(inst,"Quit")){ //if we want to quit:
            fwrite ("QUIT" , 1,  4 , socketFile ); //tell the server
            fflush(socketFile);
            fclose(socketFile);//then quit
            exit(0);
        }
        else{ //otherwise
            printf("Bad input! Try again!\n"); //there was bad input
            tryAgain = TRUE; //so set the error
        }
    }

    //format the play string and write it to the server
    sprintf(playString,"PLAY%03d",playValue);
    fwrite (playString , 1,  7 , socketFile );
    fflush(socketFile);
    return;

}

/***********************************************************/
/* Function: cleanCRLF                                     */
/* Purpose: remove the new line from a string              */
/* Parameters: char *input - the string to clean           */
/* Returns: void (none)                                    */
/***********************************************************/
void cleanCRLF(char *input){
    int stringWalker; //the walking pointer
    //for every value in the string
    for (stringWalker = 0; stringWalker < strlen(input); ++stringWalker){
        if (input[stringWalker] == '\n' ){ //if there is a new line
            input[stringWalker] = '\0'; //set it to null
            return; //get out of here
        }
    }
}

/***********************************************************/
/* Function: connectToServer                               */
/* Purpose: connects to a server on a spec'ed port         */
/* Parameters: char *hostess - the server to connect to    */
/*             int portnum - the portnumber to connect on  */
/* Returns: int - the socket file descriptor               */
/***********************************************************/
int connectToServer(char *hostess, int portnum){
    int sock; //the file descriptor
    struct sockaddr_in serverAddress; //the server address struct
    struct hostent *hostEntity; //the host struct

    sock = socket( AF_INET, SOCK_STREAM, 0 ); //get the socket
    if ( sock == -1 ) oops("socket"); //or error

    memset(&serverAddress,'\0',sizeof(serverAddress));//pre-clean the struct
    //bzero( &serverAddress, sizeof(serverAddress) );
    hostEntity = gethostbyname(hostess); //get the host name or error
    if (hostEntity == NULL) oops("location");

    bcopy(hostEntity->h_addr, (struct sockaddr *)&serverAddress.sin_addr,
            hostEntity->h_length); //copy the host address into the host entity

    //build the server address struct
    serverAddress.sin_port = htons(portnum);
    serverAddress.sin_family = AF_INET ;

    if (connect(sock,(struct sockaddr *)&serverAddress,
        sizeof(serverAddress)) !=0)  oops("connect");//connect to the socket

    return sock;//return the socket file descriptor
}
